#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <netinet/in.h>
#include <netdb.h>
#include "globals.h"
#include "util.h"

static CHAR *mapTypes = "DBbEeYmdHMSCc";

CHAR *baseName (CHAR *path)
{
    CHAR *name, *ctx, *slash;
    INT len;

    if ((name = zstrtok(path, "\t ", &ctx)) == NULL) {
        name = (CHAR *)malloc(1);
        *name = '\0';
        return name;
    }

    if ((slash = strrchr(name, '/')) == NULL)
        return name;
 
    if (!(len = strlen(slash+1))) {
        *name = '\0';
    }   
    else {
        memmove (name, slash+1, len);
        *(name+len) = '\0';
    }   

    return name;
}

CHAR *dirName (CHAR *path)
{
    CHAR *name, *ctx, *slash;

    if ((name = zstrtok(path, "\t ", &ctx)) == NULL) {
        name = (CHAR *)malloc(1);
        *name = '\0';
        return name;
    }

    if ((slash = strrchr(name, '/')) == NULL)
        *name = '\0';
    else
        *(slash+1) = '\0';

    return name;
}

/*
	take a root path and a partial (relative) path, 
        concatenate them and remove self-relative constructs ( . and ..)
*/

CHAR *realPath (CHAR *root, CHAR *relpath)
{
    CHAR *new, *in, *out, *sl, *st[3];
    INT gs, i;

    st[0] = root;
    st[1] = "/";
    st[2] = relpath;

    new = (CHAR *) malloc (strlen (root) + strlen (relpath) + 2);

    if (new != NULL) {

        out = sl = new;
        gs = 1;

        for (i = 0; i < 3; i++) {

            for (in = st[i]; *in != '\0'; in++) {

                switch (*in) {

                    case '/':
                        if (gs && in != st[0])	/* remove multiple slashes */
                            continue;
                        sl = out;
                        gs = 1;
                        break;

                    case '.':
                        if (!gs)
                            break;
                        if (*(in + 1) == '/') {	/* remove self relative */
                            in++;
                            continue;
                        }

                        if (*(in + 1) == '.')
                            if (*(in + 2) == '/') {	/* remove backup */
                                in += 2;
                                for (out = sl; out > new; out--)
                                    if (*(sl = out - 1) == '/')
                                        break;
                                continue;
                            }

                    default:
                        gs = 0;
                        break;

                }
                *out++ = *in;
            }
        }
        *out++ = '\0';
        new = realloc (new, (size_t) (out - new));
    }
    return (new);
}

/*
	get stat of parent directory of 'filename'
*/

INT dpstat (CHAR *filename, struct stat *sb)
{
    CHAR *ptr;
    INT  ios;

    if ((ptr = strrchr(filename, '/')) == NULL)
        filename = ".";
    else
        *ptr = '\0';

    sb->st_mode = 0;

    ios = stat(filename, sb);

    if (ios < 0)
        return -1;

    if (!S_ISDIR(sb->st_mode))
        return -1;

    return 0;
}


/*
	Multi-thread "strtok" (identical behaviour) but no static
	storage. All context necessary to deal with a thread is stored in
	the "context" pointer.

	additionally, there is a quoting parameter list, that defines
	quoting character pairs. Text between one of these pairs is
	excepted from the separator matching.
*/
CHAR *qstrtok (CHAR *buf, CHAR *pat, CHAR **context, CHAR *quotes)
{
    CHAR *ptr = NULL, *result;

    if (context  == NULL) return (NULL);
    if (*pat == '\0') return (*context=NULL);
    if (buf != NULL) *context = buf;
    if ((result = *context) == NULL) return (NULL);

    for (; *result != '\0'; result++)
	if (strchr(pat,(int)*result) == NULL) break;

    if (*result == '\0') return (*context=NULL);

    for (; *quotes != '\0'; quotes++)
	if (*result == *quotes) {
            if (*(++quotes) != '\0') {
                if (*(++result) == '\0') return (*context=NULL);
                    ptr = strchr(result, (int)(*quotes));
            }
            break;
        }

    if (*quotes == '\0')
        ptr = strpbrk(result, pat);

    if (ptr != NULL) *ptr++ = '\0';
    *context = ptr;

    return(result);        
}

/*
	Basically the same a "xstrtok", but doesn't modify the contents of
	the input buffer. Instead each aquired token is copied to dynamically
	allocated buffers (by using "malloc"). The caller is responsible
	for freeing these buffers if needed.
*/
CHAR *zstrtok (CHAR *buf, CHAR *pat, CHAR **context)
{
    CHAR *ptr, *temp, *result;
    INT len;

    if (context  == NULL) return (NULL);
    if (*pat == '\0') return (*context=NULL);
    if (buf != NULL) *context = buf;
    if ((temp = *context) == NULL) return (NULL);

    for (; *temp != '\0'; temp++)
	if (strchr(pat,(INT)*temp) == NULL) break;
    if (*temp == '\0') return (*context=NULL);

    ptr = strpbrk(temp, pat);

    if (ptr == NULL) {
	len = strlen(temp);
	*context = NULL;
    }
    else {
	len = (INT)(ptr - temp);
	*context = (ptr+1);
    }

    result = (CHAR *)malloc(len+1);

    if (result == NULL)
	*context = NULL;
    else {
	memcpy(result,temp,len);
	*(result+len) = '\0';
    }

    return(result);        
}


INT checkFileMap (CHAR *map)
{
    for (; *map != '\0'; map++)
        if (*map == '%')
            if (strchr(mapTypes, *(++map)) == NULL)
                return 1;
    return 0;
}


CHAR *xlateFileMap (CHAR *path, CHAR *map, struct sockaddr_in *client)
{
#define MINLEN 64

    CHAR *new = NULL, *mem, *start, *next, buffer[8], *dot;
    INT  size = 0, rest = 0, len;
    time_t now;
    struct tm lt;
    struct hostent *hp;

    now = time(NULL);
    memcpy(&lt, localtime(&now), sizeof(struct tm));

    for (; *map != '\0'; map++) {
        mem = NULL;
        if (*map == '%' && *(map+1) != '\0') {
            len = 0;
            switch (*(++map)) {
                case 'D':
                    mem = start = dirName(path);
                    if (start != NULL)
                        len = strlen(start);
                    break;

                case 'B':
                    mem = start = baseName(path);
                    if (start != NULL)
                        len = strlen(start);
                    break;

                case 'b':
                    mem = start = baseName(path);
                    if (start != NULL) {
                        if ((dot = strchr(start,'.')) != NULL)
                            *dot = '\0';
                        len = strlen(start);
                    }
                    break;

                case 'E'|'e':
                    mem = start = baseName(path);
                    if (start != NULL) {
                        if ((start = strchr(start,'.')) != NULL)
                            len = strlen(++start);
                    }
                    break;

                case 'Y':
                    sprintf(buffer, "%4.4d", lt.tm_year+1900);
                    start = buffer;
                    len = 4;
                    break;

                case 'm':
                    sprintf(buffer, "%2.2d", lt.tm_mon+1);
                    start = buffer;
                    len = 2;
                    break;

                case 'd':
                    sprintf(buffer, "%2.2d", lt.tm_mday);
                    start = buffer;
                    len = 2;
                    break;

                case 'H':
                    sprintf(buffer, "%2.2d", lt.tm_hour);
                    start = buffer;
                    len = 2;
                    break;

                case 'M':
                    sprintf(buffer, "%2.2d", lt.tm_min);
                    start = buffer;
                    len = 2;
                    break;

                case 'S':
                    sprintf(buffer, "%2.2d", lt.tm_sec);
                    start = buffer;
                    len = 2;
                    break;

                case 'C'|'c':
                    if (client != NULL) {
                        if (noResolv)
                            hp = NULL;
                        else
                            hp = gethostbyaddr((CHAR *)&client->sin_addr.s_addr,
                                               sizeof(in_addr_t),AF_INET);
                        if (hp == NULL)
                            start = inet_ntoa(client->sin_addr);
                        else {
                            if (*map == 'C') {
                                start = hp->h_name;
                            }
                            else {
                                CHAR *dot;
                                if ((dot = strrchr(hp->h_name, '.')) != NULL)
                                    start = dot+1;
                                else
                                    start = hp->h_name;
                            }
                        }
                        if (start != NULL) {
                            len = strlen(start);
                            break;
                        }
                    }

                default:
                    start = map-1;
                    len = 2;
                    break;
            }
        }
        else {
            start = map;
            len = 1;
        }
        if (len) {
            while (rest <= len) {
                INT olds;
                olds = size;
                size += ((len/MINLEN)+1)*MINLEN;
                if ((new = (CHAR *)realloc(new, size)) == NULL)
                    return NULL;
                rest += (size - olds);
                next = new + size - rest;
            }
            memcpy(next, start, len);
            rest -= len;
            next += len;
        }
        if (mem != NULL)
            free (mem);
    }
    *next = '\0';
    return new;
#undef MINLEN
}    


CHAR *fmtstat (struct stat *st)
{
    static char fmtbuf[64];
    struct passwd *pp;
    struct group *gp;
    char *pw_name, *gr_name;

    if((gp = getgrgid(st->st_gid)) == NULL)
      gr_name = "?";
    else
      gr_name = gp->gr_name;

    if((pp = getpwuid(st->st_uid)) == NULL)
      pw_name = "?";
    else
      pw_name = pp->pw_name;

    sprintf (fmtbuf,"%3.3o %s:%s", st->st_mode & 0x1ff, pw_name, gr_name);

    return fmtbuf;
}

